//===========================================================================
/*
 CS277 - Experimental Haptics
 Spring 2014, Stanford University
 
 You may use this program as a boilerplate for starting assignment #3.
 
 Additional files containing a custom subclass of cMultiMesh that loads
 and visually renders unstructured point data are provided. You will like
 write most of your code in the PointSet class, or another class you create
 and add to the project, for this homework assignment.
 
 \author    Francois Conti & Sonny Chan
 \date      January 2010, April 2014
 */
//===========================================================================

//------------------------------------------------------------------------------
#include "chai3d.h"
#include "PointSet.h"
//------------------------------------------------------------------------------
using namespace chai3d;
using namespace std;
//------------------------------------------------------------------------------
#ifndef MACOSX
#include "GL/glut.h"
#else
#include "GLUT/glut.h"
#endif
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// GENERAL SETTINGS
//------------------------------------------------------------------------------

// stereo Mode
/*
 C_STEREO_DISABLED:            Stereo is disabled
 C_STEREO_ACTIVE:              Active stereo for OpenGL NVDIA QUADRO cards
 C_STEREO_PASSIVE_LEFT_RIGHT:  Passive stereo where L/R images are rendered next to each other
 C_STEREO_PASSIVE_TOP_BOTTOM:  Passive stereo where L/R images are rendered above each other
 */
cStereoMode stereoMode = C_STEREO_DISABLED;

// fullscreen mode
bool fullscreen = false;


//------------------------------------------------------------------------------
// DECLARED VARIABLES
//------------------------------------------------------------------------------

// a world that contains all objects of the virtual environment
cWorld* world;

// a camera to render the world in the window display
cCamera* camera;

// a light source to illuminate the objects in the world
cSpotLight *light;

// a haptic device handler
cHapticDeviceHandler* handler;

// a pointer to the current haptic device
cGenericHapticDevicePtr hapticDevice;

// a virtual tool representing the haptic device in the scene
cToolCursor* tool;

// a label to display the rate [Hz] at which the simulation is running
cLabel* labelHapticRate;

// indicates if the haptic simulation currently running
bool simulationRunning = false;

// indicates if the haptic simulation has terminated
bool simulationFinished = true;

// frequency counter to measure the simulation haptic rate
cFrequencyCounter frequencyCounter;
cFrequencyCounter fpsCounter;

// information about computer screen and GLUT display window
int screenW;
int screenH;
int windowW;
int windowH;
int windowPosX;
int windowPosY;


//------------------------------------------------------------------------------
// DECLARED FUNCTIONS
//------------------------------------------------------------------------------

// callback when the window display is resized
void resizeWindow(int w, int h);

// callback when a key is pressed
void keySelect(unsigned char key, int x, int y);

// callback to render graphic scene
void updateGraphics(void);

// callback of GLUT timer
void graphicsTimer(int data);

// function that closes the application
void close(void);

// main haptics simulation loop
void updateHaptics(void);


//===========================================================================
/*
 Here is the main application entry point. This program follows the same
 structure as the CHAI3D example programs.
 
 The scene graph in this template program contains one primary object of
 type PointSet, which loads the point data from the file p1-ledge.ply.
 If the file fails to load, ensure that the paths are set correctly, and
 change them if needed. The implementation of the force rendering algorithm
 must be completed in the PointSet class before the object can be touched.
 */
//===========================================================================

int main(int argc, char* argv[])
{
    //--------------------------------------------------------------------------
    // INITIALIZATION
    //--------------------------------------------------------------------------
    
    cout << endl;
    cout << "-----------------------------------" << endl;
    cout << "CS277 - Experimental Haptics" << endl;
    cout << "Assignment #3: Point Set Surfaces" << endl;
    cout << "Spring 2014, Stanford University" << endl;
    cout << "Copyright 2003-2014" << endl;
    cout << "-----------------------------------" << endl << endl << endl;
    cout << "Keyboard Options:" << endl << endl;
    cout << "[f] - Enable/Disable full screen mode" << endl;
    cout << "[x] - Exit application" << endl;
    cout << endl << endl;
    
    
    //--------------------------------------------------------------------------
    // OPEN GL - WINDOW DISPLAY
    //--------------------------------------------------------------------------
    
    // initialize GLUT
    glutInit(&argc, argv);
    
    // retrieve  resolution of computer display and position window accordingly
    screenW = glutGet(GLUT_SCREEN_WIDTH);
    screenH = glutGet(GLUT_SCREEN_HEIGHT);
    windowW = 0.8 * screenH;
    windowH = 0.5 * screenH;
    windowPosY = (screenH - windowH) / 2;
    windowPosX = windowPosY;
    
    // initialize the OpenGL GLUT window
    glutInitWindowPosition(windowPosX, windowPosY);
    glutInitWindowSize(windowW, windowH);
    if (stereoMode == C_STEREO_ACTIVE)
    {
        glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE | GLUT_STEREO);
    }
    else
    {
        glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
    }
    
    // create display context and initialize GLEW library
    glutCreateWindow(argv[0]);
    glewInit();
    
    // setup GLUT options
    glutDisplayFunc(updateGraphics);
    glutKeyboardFunc(keySelect);
    glutReshapeFunc(resizeWindow);
    glutSetWindowTitle("CHAI3D");
    
    // set fullscreen mode
    if (fullscreen)
    {
        glutFullScreen();
    }
    
    
    //--------------------------------------------------------------------------
    // WORLD - CAMERA - LIGHTING
    //--------------------------------------------------------------------------
    
    // create a new world.
    world = new cWorld();
    
    // set the background color of the environment
    world->m_backgroundColor.setWhite();
    
    // create a camera and insert it into the virtual world
    camera = new cCamera(world);
    world->addChild(camera);
    
    // position and orient the camera
    camera->set(cVector3d (3.0, 0.0, 1.0),    // camera position (eye)
                cVector3d (0.0, 0.0, 0.0),    // lookat position (target)
                cVector3d (0.0, 0.0, 1.0));   // direction of the (up) vector
    
    // set the near and far clipping planes of the camera
    // anything in front/behind these clipping planes will not be rendered
    camera->setClippingPlanes(0.01, 10.0);
    
    // set stereo mode
    camera->setStereoMode(stereoMode);
    
    // set stereo eye separation and focal length (applies only if stereo is enabled)
    camera->setStereoEyeSeparation(0.03);
    camera->setStereoFocalLength(1.8);
    
    camera->setUseMultipassTransparency(true);
    
    // create a light source
    light = new cSpotLight(world);
    
    // attach light to camera
    camera->addChild(light);
    
    // enable light source
    light->setEnabled(true);
    
    // position the light source
    light->setLocalPos( 0.0, 0.5, 1.0);
    
    // define the direction of the light beam
    light->setDir(-2.5,-0.8, 0.0);
    
    // set light cone half angle
    light->setCutOffAngleDeg(45);
    
    
    //--------------------------------------------------------------------------
    // HAPTIC DEVICES / TOOLS
    //--------------------------------------------------------------------------
    
    // create a haptic device handler
    handler = new cHapticDeviceHandler();
    
    // get access to the first available haptic device
    handler->getDevice(hapticDevice, 0);
    
    // retrieve information about the current haptic device
    cHapticDeviceInfo hapticDeviceInfo = hapticDevice->getSpecifications();
    
    // if the haptic devices carries a gripper, enable it to behave like a user switch
    hapticDevice->setEnableGripperUserSwitch(true);
    
    // create a tool (cursor) and insert into the world
    tool = new cToolCursor(world);
    world->addChild(tool);
    
    // connect the haptic device to the tool
    tool->setHapticDevice(hapticDevice);
    
    // map the physical workspace of the haptic device to a larger virtual workspace.
    tool->setWorkspaceRadius(2.5);
    
    // define the radius of the tool (sphere)
    double toolRadius = 0.05;
    
    // define a radius for the tool
    tool->setRadius(toolRadius);
    
    // hide the device sphere. only show proxy.
    tool->setShowContactPoints(true, false);
    
    // enable if objects in the scene are going to rotate of translate
    // or possibly collide against the tool. If the environment
    // is entirely static, you can set this parameter to "false"
    tool->enableDynamicObjects(true);
    
    // start the haptic tool
    tool->start();
    
    
    //--------------------------------------------------------------------------
    // CREATE OBJECTS
    //--------------------------------------------------------------------------
    
    // read the scale factor between the physical workspace of the haptic
    // device and the virtual workspace defined for the tool
    double workspaceScaleFactor = tool->getWorkspaceScaleFactor();
    
    // stiffness properties
    double maxStiffness	= hapticDeviceInfo.m_maxLinearStiffness / workspaceScaleFactor;
    
    
    /////////////////////////////////////////////////////////////////////////
    // IMPLICIT SURFACE OBJECT(S)
    /////////////////////////////////////////////////////////////////////////

    // create the object representing the implicit surface
    PointSet *object = new PointSet();
    
    bool result = object->loadFromFile("p1-ledge.ply");
    
    // the surface effect renders a spring force between the device and proxy
    // points with the given stiffness in the material
    object->addEffect(new cEffectSurface(object));
    object->m_material->setStiffness(0.5 * maxStiffness);
    
    world->addChild(object);
    
    
    //--------------------------------------------------------------------------
    // WIDGETS
    //--------------------------------------------------------------------------
    
    // create a font
    cFont *font = NEW_CFONTCALIBRI20();
    
    // create a label to display the haptic rate of the simulation
    labelHapticRate = new cLabel(font);
    labelHapticRate->m_fontColor.setBlack();
    camera->m_frontLayer->addChild(labelHapticRate);
    
    // create a background
    cBackground* background = new cBackground();
    camera->m_backLayer->addChild(background);
    
    // set background properties
    background->setCornerColors(cColorf(1.0, 1.0, 1.0),
                                cColorf(1.0, 1.0, 1.0),
                                cColorf(0.8, 0.8, 0.8),
                                cColorf(0.8, 0.8, 0.8));
    
    
    //--------------------------------------------------------------------------
    // START SIMULATION
    //--------------------------------------------------------------------------
    
    // create a thread which starts the main haptics rendering loop
    cThread* hapticsThread = new cThread();
    hapticsThread->start(updateHaptics, CTHREAD_PRIORITY_HAPTICS);
    
    // start the main graphics rendering loop
    glutTimerFunc(50, graphicsTimer, 0);
    glutMainLoop();
    
    // close everything
    close();
    
    // exit
    return (0);
}

//------------------------------------------------------------------------------

void resizeWindow(int w, int h)
{
    windowW = w;
    windowH = h;
}

//------------------------------------------------------------------------------

void keySelect(unsigned char key, int x, int y)
{
    // option ESC: exit
    if ((key == 27) || (key == 'x'))
    {
        // close everything
        close();
        
        // exit application
        exit(0);
    }
    
    // option f: toggle fullscreen
    if (key == 'f')
    {
        if (fullscreen)
        {
            windowPosX = glutGet(GLUT_INIT_WINDOW_X);
            windowPosY = glutGet(GLUT_INIT_WINDOW_Y);
            windowW = glutGet(GLUT_INIT_WINDOW_WIDTH);
            windowH = glutGet(GLUT_INIT_WINDOW_HEIGHT);
            glutPositionWindow(windowPosX, windowPosY);
            glutReshapeWindow(windowW, windowH);
            fullscreen = false;
        }
        else
        {
            glutFullScreen();
            fullscreen = true;
        }
    }
}

//------------------------------------------------------------------------------

void close(void)
{
    // stop the simulation
    simulationRunning = false;
    
    // wait for graphics and haptics loops to terminate
    while (!simulationFinished) { cSleepMs(100); }
    
    // close haptic device
    tool->stop();
}

//------------------------------------------------------------------------------

void graphicsTimer(int data)
{
    if (simulationRunning)
    {
        fpsCounter.signal();
        glutPostRedisplay();
    }
    
    glutTimerFunc(50, graphicsTimer, 0);
}

//------------------------------------------------------------------------------

void updateGraphics(void)
{
    /////////////////////////////////////////////////////////////////////
    // UPDATE WIDGETS
    /////////////////////////////////////////////////////////////////////
    
    // update haptic rate label
    labelHapticRate->setString(
        "haptic rate: " + cStr(frequencyCounter.getFrequency(), 0) + " Hz, " +
        "visual rate: " + cStr(fpsCounter.getFrequency(), 0) + " fps");
    
    // update position of label
    labelHapticRate->setLocalPos((int)(0.5 * (windowW - labelHapticRate->getWidth())), 15);
    
    
    /////////////////////////////////////////////////////////////////////
    // RENDER SCENE
    /////////////////////////////////////////////////////////////////////
    
    // render world
    camera->renderView(windowW, windowH);
    
    // swap buffers
    glutSwapBuffers();
    
    // check for any OpenGL errors
    GLenum err = glGetError();
    if (err != GL_NO_ERROR) cout << "Error: " << gluErrorString(err) << endl;
}

//------------------------------------------------------------------------------

void updateHaptics(void)
{
    // simulation in now running
    simulationRunning  = true;
    simulationFinished = false;
    
    // main haptic simulation loop
    while(simulationRunning)
    {
        // update frequency counter
        frequencyCounter.signal(1);
        
        // compute global reference frames for each object
        world->computeGlobalPositions(true);
        
        // update position and orientation of tool
        tool->updatePose();
        
        // compute interaction forces
        tool->computeInteractionForces();
        
        // send forces to haptic device
        tool->applyForces();  
    }
    
    // exit haptics thread
    simulationFinished = true;
}

//------------------------------------------------------------------------------
